local class = require "yale/class"
local U = require "yale/util"
local SM = require "yale/soundmanager"

local Ex = require "explosion"
local Bul = require "bullet"

local abs = math.abs
local atan2 = math.atan2
local cos = memcos
local sin = memsin
local pi = math.pi

local isKeyDown = love.keyboard.isDown
local getJoyAxis = U.mild_curry(love.joystick.getAxis)(0)

local getAxis = function(axis)
	local neg, pos
	if axis == 0 then
		neg, pos = isKeyDown "left", isKeyDown "right"
	elseif axis == 1 then
		neg, pos = isKeyDown "up", isKeyDown "down"
	elseif axis == 2 then
		neg, pos = isKeyDown "a", isKeyDown "d"
	elseif axis == 3 then
		neg, pos = isKeyDown "w", isKeyDown "s"
	end
	if neg and not pos then
		return -1
	elseif pos and not neg then
		return 1
	end
	return getJoyAxis(axis)
end

local line = love.graphics.line
local polygon = love.graphics.polygon
local setColor = love.graphics.setColor
local newSoundData = love.sound.newSoundData

local function play(sd, vol)
	soundmanager:play(sd, vol)
end

local EN = require "yale/entity"

sfx_shoot = newSoundData("rsc/sfx/bullet_shoot.wav")

-- player entity
class "player" ("entity") {
	init = function(o)
		EN.init(o)
		o.max_health = 1
		o.health = o.max_health
		o.shot_stamp = 0
		o.shot_wait = .15
		o.spray_level = 0
		o.shotcycle = 0
	end
}

-- stuff
local function init_physics(p)
	p.body = p:add_body("main", 0, 0)
	p.body:setAngularDamping(32)
	p.body:setLinearDamping(10)
	p.shape = p:add_shape("main", "rectangle", "main", 0, 0, 15, 15)
	p.shape:setCategory(1) -- player category
	p.shape:setMask(3) -- player bullets category
	p:set_body_masses()
end

local function update(p, dt)
	-- are you supposed to die? if yes, please commence dying here:
	if p.health <= 0 then
		-- put body in virtual dump
		p.body:setPosition(-10000, -10000)
		p.dead = true
		-- explosion
		Ex{ world = p.world, x = p.x, y = p.y, speed = 100*p.max_health }
		return
	end
	
	-- position poll
	p.x, p.y = p.body:getPosition()
	
	-- linear movement
	local fx, fy = U.sqr2circ(getAxis(0),getAxis(1)) -- normalize input to unit circle
	fx = fx * 150 ; fy = fy * 150
	p.body:applyForce(fx, fy)
	
	-- angular stuff
	local gx = getAxis(2)
	local gy = getAxis(3)
	if abs(gx) > .5 or abs(gy) > .5 then
		-- track angle
		local theta = p.body:getAngle()
		local theta2 = atan2(gy, gx)
		local pdtheta = (theta2 - theta) % (2 * pi)
		local ndtheta = 2 * pi - pdtheta
		if pdtheta < ndtheta then
			p.body:applyTorque(pdtheta * 32)
		else
			p.body:applyTorque(-ndtheta * 32)
		end
		-- try shooting
		p:shoot(dt)
	end
end

local function draw(p)
	local hrat = p.health/p.max_health
	
	setColor(U.hsl2rgb(hrat * .3, .75 - hrat*.25, hrat*.1+.15 ))
	polygon("fill", p.shape:getPoints())
	setColor(255, 255*hrat, 255*hrat)
	polygon("line", p.shape:getPoints())
	
	local theta = p.body:getAngle()
	local x, y = p.body:getX(), p.body:getY()
	local dx, dy = cos(theta), sin(theta)
	line(x, y, x+dx*7.5, y+dy*7.5)
end

local function damage(p, d)
	p.health = p.health - d
end

local function shoot(p, dt)
	local shot_stamp = p.shot_stamp + dt
	local shot_wait = p.shot_wait
	
	while shot_stamp > 0 do
		Bul {
			world = p.world,
			x = p.x,
			y = p.y,
			cat = 3,
			angle = p.body:getAngle(),
			fcolor = { 0, 128, 255 },
			I = 5,
			dam = .3
		}
		for i=1,p.spray_level do
			Bul {
				world = p.world,
				x = p.x,
				y = p.y,
				cat = 3,
				angle = p.body:getAngle()+i*.1,
				fcolor = { 0, 128, 255 },
				I = 5,
				dam = .3
			}
			Bul {
				world = p.world,
				x = p.x,
				y = p.y,
				cat = 3,
				angle = p.body:getAngle()-i*.1,
				fcolor = { 0, 128, 255 },
				I = 5,
				dam = .3
			}
		end
		--if p.shotcycle == 0 then
			play(sfx_shoot, .5)
		--end
		--p.shotcycle = (p.shotcycle + 1) % 3
		shot_stamp = shot_stamp - shot_wait
	end
	
	p.shot_stamp = shot_stamp
end

-- export
local P = class.player
P.init_physics = init_physics
P.update = update
P.draw = draw
P.damage = damage
P.shoot = shoot
return P
