local class = require "yale/class"

local newBody = love.physics.newBody
local newCircleShape = love.physics.newCircleShape
local newRectangleShape = love.physics.newRectangleShape
local newPolygonShape = love.physics.newPolygonShape

-- entity class

class "entity" (nil) {
	init = function(o)
		local world = o.world
		assert(world, "entity declared with no world to live in")
		
		-- append to world entities so world knows of entity's existence
		local wens = world.entities
		wens[#wens+1] = o
		
		-- initialize some values
		o.windex = #wens -- entity's world index
		o.bodies = {}
		o.shapes = {}
		o.invisible = o.invisible or false
		o.paused = o.paused or false
		o.x = o.x or 0
		o.y = o.y or 0
		o:init_physics()
	end
}

-- helper functions for creating entity physics
local function add_body(e, name, lx, ly, m, i)
	-- get world physics
	local ph = e.world.physics
	-- get world coordinates by translating with entity
	local gx = e.x + lx
	local gy = e.y + ly
	-- create body
	local body = newBody(ph, gx, gy, m, i)
	-- let entity know and return
	e.bodies[name] = body
	return body
end
local function add_shape(e, name, t, bname, ...)
	-- get body
	local body = e.bodies[bname]
	assert(body, ("could not retrieve body `%s' to attach shape `%s' to"):format(bname, name))
	-- make shape
	local shape
	if t == "circle" then
		shape = newCircleShape(body, ...)
	elseif t == "rectangle" then
		shape = newRectangleShape(body, ...)
	elseif t == "polygon" then
		shape = newPolygonShape(body, ...)
	else
		error(("shape `%s' has invalid type declaration `%s'"):format(name, t))
	end
	-- set stuff and return
	e.shapes[name] = shape
	shape:setData(e)
	return shape
end
local function set_body_masses(e)
	for _,b in pairs(e.bodies) do b:setMassFromShapes() end
end

-- stuff to override in subclasses
local function init_physics(e)
	error(("override init_physics for `%s'"):format(e.type))
end
local function update(e, dt)
	error("entity must have overridden `update' for update call to work")
end
local function draw(e)
	error("entity must have overridden `draw' for draw call to work")
end

-- NOTE:
-- collision_add, collision_persist, collision_remove, and collision_result
-- can all be overridden to provide collision response to other entities

-- export
local E = class.entity
E.add_body = add_body
E.add_shape = add_shape
E.set_body_masses = set_body_masses
E.update = update
E.draw = draw
return E
