local abs = math.abs
local min = math.min
local tiny = -math.huge
local select = select
local unpack = unpack

-- various utility functions

-- mild curry
-- turns f(a, ...) into mild_curry(f)(a)(...)
local function mild_curry(f)
	return function(a)
		return function(...)
			return f(a, ...)
		end
	end
end

-- curry - it is yummy
-- turns f(a1, a2, ..., an) into curry(f, n)(a1[,|)(]a2[,|)...(]an)
local function curry(func, nargs)
	return setmetatable({ f = func , n = nargs }, {
		__call = function(t,...)
			local dots = select("#",...)
			for i=1,dots do
				local dot = select(i,...)
				dot = dot and dot or false -- false is the nil placeholder
				t[#t+1] = dot
			end
			if #t >= t.n then
				t.f(unpack(t))
				return nil
			end
			return t
		end
	})
end

-- memoization
local function memoize(func)
	return setmetatable({},
	{
		__mode = "v",
		__index = function(t, k)
			local val = func(k)
			t[k] = val
			return val
		end,
		__call = function(t, k)
			return t[k]
		end
	})
end

-- color conversion
local function rgb2hsl(r, g, b, a)
	local r, g, b = r / 256, g / 256, b / 256
	local m = min(r, g, b)
	local i, M = max_sel(r, g, b)
	local C = M - m
	local l = .5 * (M + m)
	local h,s
	if C == 0 then
		h = 0
		s = 0
	else
		if i == 1 then
			h = (g-b) / C % 6
		elseif i == 2 then
			h = (b-r) / C + 2
		elseif i == 3 then
			h = (r-g) / C + 4
		end
		h = h / 6
		
		if l <= 1/2 then
			s = C / (2 * l)
		else
			s = C / (2 - 2 * l)
		end
	end
	return h,s,l,a
end

-- and color conversion the other way
local function hsl2rgb(h, s, l, a)
	if s == 0 then
		return l, l, l
	end
	
	local C
	if l <= 1/2 then
		C = 2 * l * s
	else
		C = (2 - 2 * l) * s
	end
	
	local h = h * 6
	local X = C * (1 - abs(h % 2 - 1))
	local m = l - .5 * C
	local r, g, b
	
	if h < 1 then		r = C+m ; g = X+m ; b = m
	elseif h < 2 then	g = C+m ; r = X+m ; b = m
	elseif h < 3 then	g = C+m ; b = X+m ; r = m
	elseif h < 4 then	b = C+m ; g = X+m ; r = m
	elseif h < 5 then	b = C+m ; r = X+m ; g = m
	else				r = C+m ; b = X+m ; g = m
	end
	
	return r * 256, g * 256, b * 256, a
end

-- conformal square to circle map
-- specifically, a nice map [-1,1]^2 -> B_1(0,0)
local function sqr2circ(x, y)
	return x*((1-.5*y*y)^.5), y*((1-.5*x*x)^.5)
end

-- normalize 2d vectors
local function normalize(x, y)
	local n2 = x*x + y*y
	if n2 == 0 then return 0, 0 end
	local invn = 1/(n2^.5)
	return x*invn, y*invn
end

-- fin
local U = {}
U.mild_curry = mild_curry
U.curry = curry
U.memoize = memoize
U.rgb2hsl = rgb2hsl
U.hsl2rgb = hsl2rgb
U.sqr2circ = sqr2circ
U.normalize = normalize
return U